from singleton_tools import SingletonMeta
from typing import Dict, Union, Tuple, List, Literal
from abc import ABC, abstractmethod
from qiskit import QuantumCircuit
from info_tools import InfoTools

import numpy as np
import warnings
import random
import math



class IndividualCache(metaclass=SingletonMeta):

    def __init__(self):
        """
        Clase para dar persistencia a ciertas características de los individuos
        """

        # -- Id del individuo que se incrementa cada vez que se crea uno uno
        self.individual_id: int = 1


class IndividualMethods(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def create_individual(self, bounds_dict: Dict[str, Tuple[Union[int, float]]], child_values: List | None) -> Dict[str, Union[int, float]]:
        pass

    @abstractmethod
    def exists_malformation(self, individual_values: Dict[str, Union[int, float]], bounds_dict: Dict[str, Tuple[Union[int, float]]]) -> bool:
        pass


class BoundRestrictedIndividualMethods(IndividualMethods):
    def __init__(self):
        super().__init__()

        self.IC: IndividualCache = IndividualCache()
        self.IT: InfoTools = InfoTools()

    def exists_malformation(self, individual_values: Dict[str, Union[int, float]], bounds_dict: Dict[str, Tuple[Union[int, float]]]) -> bool:
        """
        Metodo para saber si el individuo tiene valores fuera del rango
        :return: True si existe malformacion, False else
        """

        for k, v in individual_values.items():
            individual_value: int | float = individual_values[k]
            individual_restrictions: tuple = bounds_dict[k]["malformation_limits"]

            if individual_value < min(individual_restrictions) or individual_value > max(individual_restrictions):
                return True

        return False

    def create_individual(self, bounds_dict: Dict[str, Tuple[Union[int, float]]], child_values: List | None) -> Dict[str, Union[int, float]]:

        # -- Calculamos la cantidad de qubits necesarios para cada propiedad del individuo y las almacenamos en un dict
        individual_values: Dict[str, Union[int, float]] = {}
        # -- En caso de que no se le pasen los child_list de la generacion, se crean aleatoriamente los valores
        if child_values is None:

            for parameter, v in bounds_dict.items():
                if v["bound_type"] == "interval":
                    match v["type"]:
                        case "int":
                            individual_values[parameter] = int(self.generate_random_value((v["limits"][0], v["limits"][1]), v["type"]))
                        case "float":
                            individual_values[parameter] = float(self.generate_random_value((v["limits"][0], v["limits"][1]), v["type"]))

                else:
                    individual_values[parameter] = self.generate_possible_value(v["limits"], v["type"])
        else:
            for parameter, cv in zip([z for z in bounds_dict.keys()], child_values):
                if bounds_dict[parameter]["type"] == "int":
                    individual_values[parameter] = int(cv)
                else:
                    individual_values[parameter] = float(cv)

        return individual_values


    @staticmethod
    def generate_random_value(val_tuple: tuple, data_type: str):
        if data_type == "int":
            return random.randint(val_tuple[0], val_tuple[1])
        elif data_type == "float":
            return random.uniform(val_tuple[0], val_tuple[1])

    @staticmethod
    def generate_possible_value(val_tuple: tuple, data_type: str):
        if data_type == "int":
            return random.choice(val_tuple)
        elif data_type == "float":
            return random.choice(val_tuple)


class FullRestrictedIndividualMethods(IndividualMethods):

    def __init__(self):
        super().__init__()

    def create_individual(self, bounds_dict: Dict[str, Tuple[Union[int, float]]], child_values: List | None) -> Dict[str, Union[int, float]]:
        individual_values: Dict[str, Union[int, float]] = {}
        # -- En caso de que no se le pasen los child_list de la generacion, se crean aleatoriamente los valores
        if child_values is None:

            for parameter, v in bounds_dict.items():
                individual_values[parameter] = self.generate_possible_value(v["limits"], individual_values)
        else:
            for parameter, cv in zip([z for z in bounds_dict.keys()], child_values):
                individual_values[parameter] = cv

        return individual_values

    def exists_malformation(self, individual_values: Dict[str, Union[int, float]], bounds_dict: Dict[str, Tuple[Union[int, float]]]) -> bool:
        """
        Metodo para saber si el individuo tiene valores fuera del rango
        :return: True si existe malformacion, False else
        """

        for k, v in individual_values.items():
            individual_value: int | float = individual_values[k]
            individual_restrictions: tuple = bounds_dict[k]["malformation_limits"]

            if individual_value < min(individual_restrictions) or individual_value > max(individual_restrictions):
                return True
            if len(list(individual_values.values())) > len(set(individual_values.values())):
                return True
        return False

    @staticmethod
    def generate_possible_value(val_tuple: tuple, individual_values):
        return random.choice([z for z in val_tuple if z not in individual_values.values()])


class Individual:
    def __init__(self, bounds_dict: Dict[str, Tuple[Union[int, float]]], child_values: np.ndarray | None,
                 vqc: QuantumCircuit | None , vqc_parameters: Dict[str, List[float]] | None, generation: int,
                 problem_restrictions: Literal['bound_restricted', 'totally_restricted'],
                 element_matrix: Dict[str, tuple] | None = None):
        """
        Clase que va a instanciar los distintos individuos que van a competir.
        :param bounds_dict: [str, Tuple[Union[int, float]]] Diccionario en el que se definen los parámetros a optimizar y sus valores, ej. '{learning_rate: (0.0001, 0.1)}'
        de uso y que desemboca en un individuo que se deshechará por tener una malformación. Por ejemplo, si estamos optimizando un learning_rate y la mutación nos da un valor
        superior a 1, ese individuo, se descarta antes de ser evaluado. ej. '{learning_rate: (0.000001, 1)}', si los supera, consideramos malformación.
        :param child_values: [np.ndarray | None] Valores a partir de los cuales generar los nuevos individuos.
        :param vqc: Circuito cuántico perteneciente al individuo.
        :param vqc_parameters: Dict[str, List[float]] Diccionario de parámetros para cada propiedad del individuo. Opcional para la primera generación.
        :param generation: [int] Numero de generación a la que pertenece el individuo.
        :param problem_restrictions: Literal['bound_restricted', 'totally_restricted'] Define si el individuo a crear requiere bounds o está completamente restringido.
        :param element_matrix: [Dict[str, tuple] | None] Matriz de distancia utilizada para los problemas de optimización combinatoria de tipo TSP.
        """

        # -- Instanciamos la clase cache de individuo para dar persistencia a ciertas propiedades
        self.IC: IndividualCache = IndividualCache()

        # -- Definimos el id del individuo
        self._id = self.IC.individual_id

        # -- Almacenamos parámetros en propiedades
        self.bounds_dict: Dict[str, Tuple[Union[int, float]]] = bounds_dict

        # -- Almacenamos los valores que se utilizarán para generar el individuo (sus valores reales)
        self.child_values: numpy.ndarray | None = child_values

        # -- Almacenamos el circuito cuántico con el que se crea el individuo
        self._vqc: QuantumCircuit | None = vqc

        # -- Alamcenamos los valores de los parámetros de la VQC que generaron las propiedades del individuo
        self._vqc_parameters: Dict[str, List[float]] | None = vqc_parameters

        # -- Almacenamos en una propiedad la generación a la que pertenece el individuo
        self.generation: int = generation

        # -- Almacenamos en una propiedad las restricciones a aplicar
        self.problem_restrictions: Literal['bound_restricted', 'totally_restricted'] = problem_restrictions

        # -- Definimos la propiedad en la que almacenaré el valor que la función de coste ha tenido para este individuo
        self._individual_fitness: float | None = None

        # -- Almacenamos en IMETHODS la instancia que contiene los metodos para el tipo de problema en cuestion
        self.IMETHODS = BoundRestrictedIndividualMethods() if self.problem_restrictions == "bound_restricted" else FullRestrictedIndividualMethods()

        # -- Creamos la propiedad de valores del individuo
        self._individual_values: Dict[str, Union[int, float]] = self.IMETHODS.create_individual(self.bounds_dict, self.child_values) if self.child_values is not None else {}

        # -- Almacenamos en una propiedad si el individuo tiene una malformación
        self._malformation: bool = self.exists_malformation()

        # -- Alamcenamos una matriz de distancia
        self._element_matrix: Dict[str, tuple] | None = element_matrix

        # -- Aumentamos en uno el individual_id para el próximo individuo
        self.IC.individual_id +=1

    def __eq__(self, other, decimals=4):
        """
        Compara si dos individuos son iguales en base a sus propiedades con precisión decimal.

        :param other: Otro objeto de la clase Individual con el que se realizará la comparación.
        :param decimals: Número de decimales a considerar en la comparación (por defecto 4).

        :return: True si ambos individuos tienen las mismas propiedades con la precisión dada, False en caso contrario.
        """

        # Verificar que el otro objeto es de la clase Individual
        if not isinstance(other, Individual):
            return False

        # Obtener los valores de los individuos
        values_self = self.get_individual_values()
        values_other = other.get_individual_values()

        # Redondear los valores antes de compararlos
        rounded_self = {k: round(v, decimals) if isinstance(v, float) else v for k, v in values_self.items()}
        rounded_other = {k: round(v, decimals) if isinstance(v, float) else v for k, v in values_other.items()}

        return rounded_self == rounded_other

    def exists_malformation(self) -> bool:
        """
        Metodo para saber si el individuo tiene valores fuera del rango
        :return: True si existe malformacion, False else
        """

        # -- NOTA: El metodo varía en funcion de las restricciones de self.problem_restrictions
        return self.IMETHODS.exists_malformation(self._individual_values, self.bounds_dict)

    # <editor-fold desc="Getters y setters    --------------------------------------------------------------------------------------------------------------------------------">

    def get_individual_id(self) -> int:
        """
        Metodo que devuelve el id del individuo
        :return:
        """
        return self._id

    def get_individual_values(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que devuelve los valores del individuo en un diccionario. Por ejemplo, si viene asi: {learning_rate: 0.0125, batch_size: 34}
        :return:
        """
        return self._individual_values

    def get_vqc(self) -> QuantumCircuit:
        """
        Metodo que devuelve el circuito cuántico con el que se crearon las propiedades del individuo.
        :return: El circuito cuántico con el que se generaron las propiedades del individuo
        """
        return self._vqc

    def get_vqc_parameters_values(self) -> Dict[str, List[float]] | None:
        """
        Metodo que devuelve los valores de los parámetros que generaron las propiedades del individuo.
        :return: Los parámetros que generaron las propiedades
        """
        return self._vqc_parameters

    def get_individual_fitness(self) -> int | float:
        """
        Metodo que devuelve el resultado de la función objetivo.
        :return: Los parámetros que generaron las propiedades
        """
        return self._individual_fitness

    def get_individual_malformation(self) -> bool:
        """
        Metodo que devuelve el resultado de la función objetivo.
        :return: Los parámetros que generaron las propiedades
        """
        return self._malformation

    def get_element_matrix(self) -> Dict[str, tuple] | None:
        """
        Metodo que devuelve el resultado de la función objetivo.
        :return: Los parámetros que generaron las propiedades
        """
        return self._element_matrix

    def add_or_update_variable(self, var_name: str, value: int | float) -> None:
        """
        Agrega o actualiza una variable de instancia en el objeto Individual.

        :param var_name: Nombre de la variable de instancia.
        :param value: Valor de la variable (puede ser de cualquier tipo).
        """
        setattr(self, f"_{var_name}", value)

    def set_individual_value(self, parameter: str, new_value: float | int):
        self._individual_values[parameter] = new_value
        self._malformation = self.exists_malformation()

    def set_individual_angles(self, parameter: str, new_value: list):
        self.get_vqc_parameters_values()[parameter] = new_value

    def set_individual_angle(self, angles: list):
        self._vqc_parameters = angles

    def set_individual_values(self, new_value_list: List[float | int]):
        for key, new_value in zip(self._individual_values.keys(), new_value_list):
            self._individual_values[key] = new_value
        self._malformation = self.exists_malformation()

    def get_child_values(self):
        return self.child_values

    # </editor-fold>

__all__ = ['IndividualCache', 'IndividualMethods', 'BoundRestrictedIndividualMethods', 'FullRestrictedIndividualMethods', 'Individual']



from abc import ABC, abstractmethod
from info_tools import InfoTools
from typing import Literal

import pandas as pd
import numpy as np



class VariabilityExplossion(ABC):
    def __init__(self, early_stopping_generations: int, problem_type: Literal['maximize', 'minimize'], round_decimals: int = 3, verbose: bool = False):
        """
        Constructor de la clase CrazyVariabilityExplossion.
        Este constructor inicializa los parámetros relacionados con el early stopping y la explosión de variabilidad,
        heredando de la clase base VariabilityExplossion.

        :param early_stopping_generations: [int] Número de generaciones que se deben esperar para activar la explosión de variabilidad si la moda se repite.
        :param problem_type: [Literal] Tipo de problema a resolver, puede ser 'maximize' o 'minimize'.
        :param round_decimals: [int] Número de decimales para redondear en las estadísticas de variabilidad.
        :param verbose: [bool] Indica si se deben mostrar mensajes detallados sobre el estado de la explosión de variabilidad.
        """

        # -- Obtenemos las generaciones que voy a esperar para que si se repite la moda, arrancar la explosion de variabilidad
        self.early_stopping_generations: int = early_stopping_generations

        # -- Almacenamos el tipo de problema e instancio InfoTools
        self.problem_type: Literal['maximize', 'minimize'] = problem_type
        self.IT: InfoTools = InfoTools()
        self.round_decimals: int = round_decimals
        self.verbose: int = verbose

        # -- Inicializamos propiedades de control de flujo
        self.early_stopping_generations_executed_counter: int = 0  # Contador de cuantas generacioes han transcurrido desde la ultima explosion de variabilidad
        self.total_early_stopping_generations_executed_counter: int = 0  # Contador de cuantas generacioes han transcurrido desde la primera explosion de variabilidad
        self.early_stopping_generations_executed: bool = False  # Boleana que indica si estamos en modo explosion de variabilidad

        # -- Inicializamos propiedades de mutacion (valores por defecto)
        self.mutate_probability: float = 0.4
        self.mutate_gen_probability: float = 0.25
        self.mutation_policy: Literal['soft', 'normal', 'hard'] = 'soft'

    @abstractmethod
    def evaluate_early_stopping(self, generations_fitness_statistics_df: pd.DataFrame | None) -> None:
        """
        Metodo que evalúa si se deben activar las condiciones para el early stopping y la explosión de variabilidad.
        Si la moda en las últimas generaciones ha sido constante, se activa la explosión de variabilidad.
        :param generations_fitness_statistics_df: [pd.DataFrame | None] DataFrame con las estadísticas de fitness de las generaciones.

        :return: tuple (m_proba, m_gen_proba, m_policy, early_stopping_generations_execute)
            - m_proba: Probabilidad de mutación de un individuo.
            - m_gen_proba: Probabilidad de mutación de un gen.
            - m_policy: Política de mutación (puede ser 'soft', 'normal', 'hard').
            - early_stopping_generations_execute: Indica si se debe ejecutar el early stopping.
        """
        pass

    @abstractmethod
    def execute_variability_explossion(self):
        """
        Metodo que ejecuta la explosión de variabilidad en caso de que el algoritmo genético haya quedado atrapado en un mínimo local.
        Este metodo incrementa las probabilidades de mutación para intentar escapar de un mínimo local y mejorar el rendimiento del algoritmo.

        :return: tuple (mutate_probability, mutate_gen_probability, mutation_policy, early_stopping_generations_executed)
            - mutate_probability: Probabilidad de mutación de un individuo.
            - mutate_gen_probability: Probabilidad de mutación de un gen.
            - mutation_policy: Política de mutación (puede ser 'soft', 'normal', 'hard').
            - early_stopping_generations_executed: Si se debe ejecutar el early stopping.
        """
        pass

    @abstractmethod
    def stop_genetic_iterations(self, generations_fitness_statistics_df: pd.DataFrame | None):
        """
        Metodo que evalúa si es necesario detener el proceso de evolución en base a las condiciones de early stopping.
        Si el algoritmo ha estado ejecutando la explosión de variabilidad y no ha mejorado el resultado en un número determinado de generaciones,
        se detiene la evolución.
        :param generations_fitness_statistics_df: [pd.DataFrame | None] DataFrame con las estadísticas de fitness de las generaciones.

        :return: [bool] Retorna True si se debe detener la evolución, False si se debe continuar.
        """
        pass

    @abstractmethod
    def print_variability_status(self):
        """
        Metodo que imprime el estado de la explosión de variabilidad.
        Muestra información sobre si la explosión está activada, cuántas generaciones lleva activa y si se ha mejorado el resultado.
        """
        pass


class CrazyVariabilityExplossion(VariabilityExplossion):
    def __init__(self, early_stopping_generations: int, problem_type: Literal['maximize', 'minimize'], round_decimals: int = 3, verbose: bool = False):

        # -- Obtengo las generaciones que voy a esperar para que si se repite la moda, arrancar la explosion de variabilidad
        super().__init__(early_stopping_generations, problem_type, round_decimals, verbose)

    def evaluate_early_stopping(self, generations_fitness_statistics_df: pd.DataFrame | None) -> tuple:
        """
        Metodo para evaluar si se han cumplido las condiciones del early stopping
        :param generations_fitness_statistics_df: [pd.Dataframe] Df con la información general de la generación
        :return:
            m_proba: Probabilidad de mutación de un individuo
            m_gen_proba: Probabilidad de mutación de un gen
            m_policy: Política de mutación
            early_stopping_generations_execute: Si se debe ejecutar el early stopping
        """

        # -- Aplicamos la explosión de variabilidad, si en las últimas early_stopping_generations el max es igual
        if generations_fitness_statistics_df is not None:
            df: pd.DataFrame = generations_fitness_statistics_df
            df_tail = df.tail(self.early_stopping_generations)
            if df.shape[0] >= int(self.early_stopping_generations * 2):
                mode_values = df_tail["min"].values if self.problem_type == "minimize" else df_tail["max"].values
                if np.all(mode_values == mode_values[0]):
                    return self.execute_variability_explossion()

        # -- Si ya se ha ejecutado la explosion de variabilidad, se vuelve a ejecutar en cada iteracion
        if self.early_stopping_generations_executed:
            return self.execute_variability_explossion()

        # -- Si no ha pasado Nada, retornamos None
        return None, None, None, None

    def execute_variability_explossion(self):
        """
        Metodo para ejecutar una explosion de variabilidad en caso de que el genético haya quedado clavado en un mínimo. Esto pretende dar una oportunidad extra al modelo genético
        de encontrar un mínimo mejor cuando ha quedado atrapado en un mínimo local.
        :return:
            mutate_probability: Probabilidad de mutación de un individuo
            mutate_gen_probability: Probabilidad de mutación de un gen
            mutation_policy: Política de mutación
            early_stopping_generations_executed: Si se debe ejecutar el early stopping
        """

        if self.early_stopping_generations_executed:
            self.mutate_probability: float = 0.4
            self.mutate_gen_probability: float = 0.25
            self.mutation_policy: Literal['soft', 'normal', 'hard'] = 'soft'
        else:
            self.mutate_probability: float = 0.9
            self.mutate_gen_probability: float = 0.5
            self.mutation_policy: Literal['soft', 'normal', 'hard'] = 'hard'
            self.early_stopping_generations_executed = True

        return self.mutate_probability, self.mutate_gen_probability, self.mutation_policy, self.early_stopping_generations_executed

    def stop_genetic_iterations(self, generations_fitness_statistics_df: pd.DataFrame | None):
        """
        Metodo para que en caso de que si se ha ejecutado el execute_variability_explossion (early_stopping_generations_executed = True), y hayan pasado n generaciones, se detiene.
        :param generations_fitness_statistics_df: [pd.Dataframe] Df con los resultados estadísticos de los individuos

        :return: Valor booleano
        """

        if self.early_stopping_generations_executed_counter >= self.early_stopping_generations:

            # -- Si no ha mejorado en las vueltas, True, sino reseteamos el contador para darle margen
            df: pd.DataFrame = generations_fitness_statistics_df
            best_value = max(df["max"].values) if self.problem_type == "maximize" else min(df["max"].values)
            best_counter_value = max(df.tail(self.early_stopping_generations_executed_counter)["max"].values) if self.problem_type == "maximize" else (
                min(df.tail(self.early_stopping_generations_executed_counter)["max"].values))

            if self.problem_type == "maximize":
                if best_counter_value <= best_value:
                    self.IT.info_print("El CRAZY MODE no ha conseguido mejorar el resultado. Detenemos el proceso de evolución", "light_yellow")
                    return True
                else:
                    self.IT.info_print("El CRAZY MODE ha mejorado el mejor resultado. Le damos margen de generaciones", "light_blue")
                    self.early_stopping_generations_executed_counter = 0
                    return False
            else:
                if best_counter_value >= best_value:
                    self.IT.info_print("El CRAZY MODE no ha conseguido mejorar el resultado. Detenemos el proceso de evolución", "light_yellow")
                    return True
                else:
                    self.IT.info_print("El CRAZY MODE ha mejorado el mejor resultado. Le damos margen de generaciones", "light_blue")
                    self.early_stopping_generations_executed_counter = 0
                    return False

        if self.early_stopping_generations_executed:
            self.early_stopping_generations_executed_counter += 1
            self.total_early_stopping_generations_executed_counter += 1

        # -- En caso de verbose activo, mostramos la info
        if self.verbose:
            self.print_variability_status()

        return False

    def print_variability_status(self):
        """
        Metodo que imprime el estado de la explosión de variabilidad.
        Muestra información sobre si la explosión está activada, cuántas generaciones lleva activa y si se ha mejorado el resultado.
        """

        self.IT.sub_intro_print(f"Resumen de CrazyVariabilityExplossion")
        self.IT.info_print(f"CrazyVariabilityExplossion Activated: {self.early_stopping_generations_executed}",
                           "light_red" if self.early_stopping_generations_executed else "light_green")
        if self.early_stopping_generations_executed:
            self.IT.info_print(f"Generaciones que lleva activo el CrazyVariabilityExplossion: {self.total_early_stopping_generations_executed_counter}")
            self.IT.info_print(f"Generaciones desde ultima mejora: {self.early_stopping_generations_executed_counter}")


from aletheia_quantum_genetic_optimizers.quantum_technology import QuantumTechnology
from aletheia_quantum_genetic_optimizers.individuals import Individual

from typing import List, Dict, Tuple, Union, Literal
from qiskit import QuantumCircuit
from info_tools import InfoTools

import plotly.graph_objects as go
import matplotlib.pyplot as plt
import scipy.stats as stats
import pandas as pd
import numpy as np
import colorsys
import math


class Population:
    def __init__(self,
                 bounds_dict: Dict[str, Tuple[Union[int, float]]],
                 num_individuals: int,
                 problem_restrictions: Literal['bound_restricted', 'totally_restricted'],
                 round_decimals: int = 3,
                 ):
        """
        Clase población
        :param bounds_dict: Diccionario de propiedades del individuo
        :param num_individuals: Numero de individuos de la población
        :param problem_restrictions: Definimos si es un problema de bounds o de restricción completa
        :param round_decimals: Redondeo de números decimales para comparar similitudes
        """

        # -- Definimos variables generales
        self.IT: InfoTools = InfoTools()
        self.bounds_dict: Dict[str, Tuple[Union[int, float]]] = bounds_dict
        self.num_individuals: int = num_individuals
        self.problem_restrictions: Literal['bound_restricted', 'totally_restricted'] = problem_restrictions
        self.populuation_dict: Dict[str, List[Individual]] = {str(0): []}
        self.generations_fitness_statistics_df: pd.DataFrame | None = None
        self.round_decimals: int = round_decimals

        # -- Definimos la variable de QuantumTechnology
        self.QT: QuantumTechnology | None = None

        # -- Definimos la variable que contendrá los parámetros del VQC
        self.vqc_parameters: Dict | None = None

        # -- Definimos el diccionario de hiperparámetros
        self.hyperparameters: dict = {key: values["type"] for key, values in self.bounds_dict.items() if key != 'objective_function_values' and key != "generation" and key != "malformation"}

        # -- Definimos la cantidad de qubits del circuito cuántico a partir de la cantidad de hiperparámetros
        self.num_qubits = len(self.hyperparameters)

    def create_population(self, quantum_technology: Literal["simulator", "quantum_machine"], service: Literal["aer", "ibm"], qm_api_key: str | None, qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None, quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"], max_qubits: int, element_matrix: Dict[str, tuple] | None = None):
        """
        Metodo para crear la población inicial
        :param quantum_technology. Literal["simulator", "quantum_machine"]. Tecnología cuántica con la que calculan la lógica. Si es simulator, se hará con un simulador definido en el
        parámetro service. Si es quantum_machine, el algoritmo se ejecutará en una máquina cuántica definida en el parámetro technology.
        :param service. ["aer", "ibm"] El servicio tecnológico con el cual se ejecuta la lógica.
        :param qm_api_key. API KEY para conectarse con el servicio de computación cuántica de una empresa.
        :param qm_connection_service. Literal["ibm_quantum", "ibm_cloud"] | None. Servicio específico de computación cuántica. Por ejemplo, en el caso de IBM pueden ser a la fecha ibm_quantum | ibm_cloud
        :param quantum_machine. Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"]. Nombre del ordenador cuántico a utilizar. Por ejemplo, en el caso de IBM puede ser ibm_brisbane, ibm_kyiv, ibm_sherbrooke. Si se deja en least_busy,
        se buscará el ordenador menos ocupado para llevar a cabo la ejecución del algoritmo cuántico.
        :param max_qubits: Máximo numero de qubits permitidos para utilizar el simulador o el ordenador cuántico.
        :param element_matrix: Dict[str, tuple] | None = Matriz de distancia utilizada para los problemas de optimización combinatoria de tipo TSP.
        :return:
        """

        # -- Instaciamos la clase que contiene las propiedades y métodos cuánticos
        self.QT: QuantumTechnology = QuantumTechnology(quantum_technology=quantum_technology,
                                                       service=service,
                                                       qm_api_key=qm_api_key,
                                                       qm_connection_service=qm_connection_service,
                                                       quantum_machine=quantum_machine,
                                                       max_qubits=max_qubits)

        match self.problem_restrictions:

            case "bound_restricted":

                # -- Creamos el circuito que generará los numeros random para los ángulos de los parámetros de valores padres
                self.IT.sub_intro_print("Creamos el circuito cuántico de cada individuo")
                vqc, parameters = self.QT.create_generator(self.bounds_dict)

                # -- Calculamos los valores de los parámetros
                self.IT.sub_intro_print("Calculamos los valores de los parámetros/ángulos de rotación")
                self.vqc_parameters: List[List[float]] = self.QT.generate_parameters(self.num_individuals, parameters)

                # -- Transformamos el formato para introducir los ángulos en los individuos cuando sea el momento
                reshaped_vqc_params = []
                num_props, keys = len(self.bounds_dict), list(self.bounds_dict.keys())
                reshaped_vqc_params.extend({key: [qc_pair[i]] for i, key in enumerate(keys)} for qc_pair in self.vqc_parameters)

                # -- Generamos las propiedades ejecutando los circuitos cuánticos con los parámetros calculados para los VQC
                self.IT.sub_intro_print("Calculamos las propiedades")
                all_properties: np.ndarray = self.QT.generate_properties(vqc, parameters, self.vqc_parameters, self.num_qubits)

                # -- Desnormalizamos los valores en el rango de limits de cada propiedad
                self.IT.sub_intro_print("Desnormalizamos los valores en el rango de limits de cada propiedad")
                properties_desnormalized: np.ndarray = self.QT.denormalize_hyperparameters(self.bounds_dict, all_properties)

                # -- Completamos la población con los individuos creados
                self.IT.sub_intro_print("Completamos la población con los individuos creados sin malformaciones")
                while len(self.populuation_dict[str(0)]) < self.num_individuals:

                    for idx, child_values in enumerate(properties_desnormalized):

                        # -- Creamos los individuos
                        individual: Individual = Individual(bounds_dict=self.bounds_dict,
                                                            child_values=child_values,
                                                            vqc=vqc.copy(),
                                                            vqc_parameters=reshaped_vqc_params[idx],
                                                            generation=0,
                                                            problem_restrictions=self.problem_restrictions)

                        # -- Si no hay malformación agregamos el individuo a la población
                        if not individual.get_individual_malformation():
                            self.populuation_dict[str(0)].append(individual)

            case "totally_restricted":

                # -- Calculamos la cantidad de qubits
                aprox_exp: float = math.log2(self.num_individuals) + 1

                exponent: int = math.floor(aprox_exp)

                # -- Validamos que el numero de qubits necesarios no sea mayor que el máximo de qubits definido
                if exponent > max_qubits:
                    raise ValueError("El número de qubits necesarios para los bounds es mayor que max_qubits")

                # -- Obtenemos el primer individuo de los bounds para extrae información general de todos los individuos
                indv_bounds = self.bounds_dict[[z for z in self.bounds_dict.keys()][0]]

                # -- Completamos la población con los individuos creados
                self.IT.sub_intro_print("Completamos la población con los individuos creados sin malformaciones)")
                while len(self.populuation_dict[str(0)]) < self.num_individuals:

                    # -- Utilizamos la computación cuántica para generar los valores de los futuros individuos
                    individual: List[int | float] = self.QT.get_permutations_hybrid(bounds_dict=indv_bounds,
                                                                                    log2=exponent,
                                                                                    shots=20000)

                    # -- Creamos los individuos
                    individual: Individual = Individual(bounds_dict=self.bounds_dict,
                                                        child_values=individual,
                                                        vqc=None,
                                                        vqc_parameters=None,
                                                        generation=0,
                                                        problem_restrictions=self.problem_restrictions,
                                                        element_matrix=element_matrix)

                    # -- Si el individuo no tiene malformaciones
                    if not individual.get_individual_malformation():

                        # -- Chequeamos que no haya otro individuo con los mismos valores
                        is_duplicate = any(existing_ind.get_individual_values() == individual.get_individual_values()
                                           for existing_ind in self.populuation_dict[str(0)])

                        # -- Apendeamos el individuo
                        if not is_duplicate:
                            self.populuation_dict[str(0)].append(individual)
                            self.IT.info_print(f"INDIVIDUO OK: {individual.get_individual_values()}", "light_yellow")

                for indv in self.populuation_dict[str(0)]:
                    self.IT.info_print(f"{indv.get_individual_values()}")

    # <editor-fold desc="Getters y setters    ----------------------------------------------------------------------------------------------------------------------------------">

    def add_generation_population(self, children_list: List[Individual], generation: int) -> None:
        """
        Metodo para agregar una generación de individuos al diccionario de población
        :param children_list:
        :param generation:
        :return: None
        """
        self.populuation_dict[str(generation)] = children_list

    def get_generation_fitness_statistics(self, generation: int) -> pd.DataFrame:
        """
            Calcula estadísticas descriptivas de una lista de valores numéricos.

            :param generation: Generación de la que vamos a obtener la info
            :return: Diccionario con media, mediana, desviación estándar, cuartiles, rango, moda y más.
            """
        values = np.array([z.get_individual_fitness() for z in self.populuation_dict[str(generation)] if not z.get_individual_malformation()])

        data = {
            "generation": [generation],
            "count": [len(values)],
            "mean": [np.mean(values)],
            "median": [np.median(values)],
            "std_dev": [np.std(values, ddof=1)],
            "variance": [np.var(values, ddof=1)],
            "min": [np.min(values)],
            "max": [np.max(values)],
            "range": [np.ptp(values)],
            "q1": [np.percentile(values, 25)],
            "q2": [np.percentile(values, 50)],
            "q3": [np.percentile(values, 75)],
            "iqr": [stats.iqr(values)],
            "mode": [stats.mode([round(v, self.round_decimals) for v in values], keepdims=True)[0][0]]
        }

        if self.generations_fitness_statistics_df is None:
            self.generations_fitness_statistics_df = pd.DataFrame(data)
        else:
            self.generations_fitness_statistics_df = pd.concat([self.generations_fitness_statistics_df, pd.DataFrame(data)], axis=0)

        return self.generations_fitness_statistics_df

    # </editor-fold>

    # <editor-fold desc="Metodos de graficación    -------------------------------------------------------------------------------------------------------------------------------">
    def plot_generation_stats(self, variability_explosion_starts_in_generation: int | None):
        fig = go.Figure()

        # Agregar líneas al gráfico
        show_stats = ['mean', 'median', 'mode', 'min', 'max', 'q1', 'q3']
        colors = ['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692']

        for stat, color in zip(show_stats, colors):
            fig.add_trace(go.Scatter(
                x=self.generations_fitness_statistics_df['generation'],
                y=self.generations_fitness_statistics_df[stat],
                mode='lines+markers',
                name=stat.capitalize(),
                line=dict(color=color, width=2, shape='spline'),  # Líneas más suaves
                marker=dict(size=6, opacity=0.8)
            ))

        # Agregar línea vertical si el valor no es None
        if variability_explosion_starts_in_generation is not None:
            x_pos = variability_explosion_starts_in_generation - 0.2  # Desplazamiento a la izquierda

            # Línea vertical
            fig.add_shape(
                type="line",
                x0=x_pos, x1=x_pos,
                y0=self.generations_fitness_statistics_df[show_stats].min().min(),
                y1=self.generations_fitness_statistics_df[show_stats].max().max(),
                line=dict(color="red", width=3, dash="dash"),
            )

            # Añadir un marcador invisible para la leyenda
            fig.add_trace(go.Scatter(
                x=[None], y=[None],  # No se mostrará en la gráfica, solo en la leyenda
                mode='lines',
                name="Variability Explosion Start",
                line=dict(color="red", width=3, dash="dash")
            ))

        # Configuración del diseño
        fig.update_layout(
            title=dict(
                text="Evolución de Estadísticas por Generación",
                font=dict(size=22, family='Arial, sans-serif', color='#2C3E50'),
                x=0.5, xanchor='center', pad=dict(b=20, t=20)
            ),
            plot_bgcolor='rgba(245,248,250,0.9)',
            paper_bgcolor='white',
            xaxis=dict(
                title="Generación",
                showgrid=True,
                zeroline=False,
                gridcolor='lightgrey',
                zerolinecolor='lightgrey',
                showline=True,
                linewidth=2,
                linecolor='#BDC3C7'
            ),
            yaxis=dict(
                title="Valor",
                showgrid=True,
                zeroline=False,
                gridcolor='lightgrey',
                zerolinecolor='lightgrey',
                showline=False,
                linewidth=2,
                linecolor='#BDC3C7'
            ),
            legend=dict(
                title='Estadísticas',
                bgcolor='rgba(255,255,255,0.9)',
                bordercolor='#BDC3C7',
                borderwidth=1,
                x=1.05, xanchor='left'
            ),
            margin=dict(l=50, r=200, t=50, b=50)
        )

        fig.show()

    def plot_evolution_animated(self, problem_type: Literal['minimize', 'maximize'] = "maximize", transition_duration_ms: int = 50) -> None:

        data_dict = self.populuation_dict
        generations = sorted(data_dict.keys())  # Obtener las generaciones en orden

        # Determinar si es un problema de minimización o maximización
        is_minimization = True if problem_type == "minimize" else False

        # Obtengo losfitness mas grandes y mas pequeños
        max_fitness_value: float = max(
            [ind.individual_fitness for ind in [elemento for sublista in data_dict.values() for elemento in sublista] if ind.individual_fitness is not None])
        min_fitness_value: float = min(
            [ind.individual_fitness for ind in [elemento for sublista in data_dict.values() for elemento in sublista] if ind.individual_fitness is not None])

        # Preparar los datos para cada fotograma de la animación
        frames = []
        max_individuals = max(len(generation) for generation in data_dict.values())

        def generate_distinct_colors(n):
            colors = []
            for i in range(n):
                hue = i / n
                saturation = 0.7
                lightness = 0.5
                rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
                colors.append(f'rgb({int(rgb[0] * 255)},{int(rgb[1] * 255)},{int(rgb[2] * 255)})')
            return colors

        # Generar un color por generación
        generation_colors = generate_distinct_colors(len(generations))
        generation_color_map = dict(zip(generations, generation_colors))

        # Crear el layout inicial con individuos de color de su generación
        initial_data = []
        for generation in generations:
            population = data_dict[generation]
            gen_color = generation_color_map[generation]
            for i, ind in enumerate(population):
                initial_data.append(go.Scatter(
                    x=[i],
                    y=[ind.individual_fitness],  # Usar el fitness real del individuo
                    mode='markers',
                    marker=dict(
                        color=gen_color,
                        size=10,
                        opacity=0.7,
                        line=dict(width=1.5, color='DarkSlateGrey')
                    ),
                    name=f'Gen {generation} Ind {i}',
                    text=[f'Gen: {generation}<br>Ind: {i}<br>Fitness: {ind.individual_fitness if ind.individual_fitness is not None else 0:.3f}'],
                    hoverinfo='text',
                    textposition='top center'
                ))

        # Preparar frames para la animación
        for generation in generations:
            # Extraer los valores de fitness de cada individuo en la generación
            population = data_dict[generation]
            fitness_values = [ind.individual_fitness for ind in population if ind.individual_fitness is not None]

            # Determinar el punto de inicio basado en minimización o maximización
            start_values = [1 if is_minimization else 0] * len(fitness_values)

            # Color de la generación actual
            gen_color = generation_color_map[generation]

            # Crear un frame que represente los valores de fitness para esta generación
            scatter_data = []

            # Scatter plot para individuos
            for i, (start, end, color) in enumerate(zip(start_values, fitness_values, [gen_color] * len(fitness_values))):
                scatter_data.append(go.Scatter(
                    x=[i],  # Posición del individuo
                    y=[start],  # Valor inicial
                    mode='markers+text',
                    marker=dict(
                        color=color,
                        size=10,
                        opacity=0.8
                    ),
                    text=[f'Gen: {generation}<br>Ind: {i}<br>Fitness: {end:.3f}'],
                    textposition='top center',
                    hoverinfo='text',
                    hoverlabel=dict(bgcolor='white', font_size=12),
                    name=f'Gen {generation} Ind {i}'
                ))

            # Frame para la animación de transición
            frame = go.Frame(
                data=scatter_data,
                name=f'Generation {generation}'
            )
            frames.append(frame)

        # Configurar la figura con animación
        fig = go.Figure(
            data=initial_data,
            layout=go.Layout(
                # Título con estilo profesional
                title=dict(
                    text='Evolución de la Función de Fitness a lo largo de las generaciones',
                    font=dict(
                        size=22,
                        family='Arial, sans-serif',
                        color='#2C3E50'  # Azul profundo elegante
                    ),
                    x=0.5,
                    xanchor='center',
                    pad=dict(b=20, t=20)
                ),

                # Fondo profesional
                plot_bgcolor='rgba(245,248,250,0.9)',  # Gris azulado muy claro
                paper_bgcolor='white',

                # Menú de animación centrado y con estilo
                updatemenus=[dict(
                    type='buttons',
                    x=0.5,  # Centrado horizontalmente
                    xanchor='center',
                    y=0.95,  # Posicionado justo debajo del título
                    buttons=[dict(
                        label='▶ Iniciar Animación',
                        method='animate',
                        args=[None, {
                            'frame': {'duration': transition_duration_ms, 'redraw': True},
                            'fromcurrent': True,
                            'transition': {
                                'duration': 5000,
                                'easing': 'cubic-in-out'
                            },
                            'mode': 'immediate'
                        }]
                    )]
                )],

                xaxis=dict(
                    title='Individuos',
                    range=[-1, max_individuals],
                    tickmode='linear',
                    tick0=0,
                    dtick=1,
                    gridcolor='lightgrey',
                    zerolinecolor='lightgrey',
                    showline=True,
                    linewidth=2,
                    linecolor='#BDC3C7'
                ),
                yaxis=dict(
                    title='Fitness',
                    range=[-0.1 if min_fitness_value > -0.1 else min_fitness_value, 1.1 if max_fitness_value < 1.1 else max_fitness_value],
                    gridcolor='lightgrey',
                    zerolinecolor='lightgrey',
                    showline=False,
                    linewidth=2,
                    linecolor='#BDC3C7'
                ),

                # Leyenda con estilo
                legend=None,
                showlegend=False,

                hovermode='closest',
            ),
            frames=frames
        )

        # Añadir fotogramas intermedios para la animación suave
        for generation in generations:
            population = data_dict[generation]
            fitness_values = [ind.individual_fitness for ind in population if ind.individual_fitness is not None]

            # Determinar el punto de inicio basado en minimización o maximización
            start_values = [(1 if max_fitness_value < 1 else max_fitness_value) if is_minimization else (0 if min_fitness_value > 0 else min_fitness_value)] * len(fitness_values)

            # Color de la generación actual
            gen_color = generation_color_map[generation]

            # Crear fotogramas de transición
            start_progress_limit: float = (1 if max_fitness_value < 1 else max_fitness_value) if is_minimization else (0 if min_fitness_value > 0 else min_fitness_value)
            end_progress_limit: float = (1 if max_fitness_value < 1 else max_fitness_value) if not is_minimization else (0 if min_fitness_value > 0 else min_fitness_value)
            animation_range: float = (end_progress_limit - start_progress_limit) / 100
            for progress in np.arange(start_progress_limit, end_progress_limit, animation_range):
                transition_scatter_data = []
                for i, (start, end, color) in enumerate(zip(start_values, fitness_values, [gen_color] * len(fitness_values))):
                    # Interpolación lineal entre el inicio y el final
                    current_y = start + progress * (end - start)
                    transition_scatter_data.append(go.Scatter(
                        x=[i],
                        y=[current_y],
                        mode='markers+text',
                        marker=dict(
                            color=color,
                            size=12,
                            opacity=0.8
                        ),
                        text=[f'<b>[Gen]</b>:<br>{generation}<br><b>[Ind]</b>:<br>{i}<br><b>[Fitness]</b><br>{end:.3f}'],
                        textposition='top center',
                        hoverinfo='text',
                        hoverlabel=dict(bgcolor='white', font_size=10),
                        name=f'Gen {generation} Ind {i}'
                    ))

                transition_frame = go.Frame(
                    data=transition_scatter_data,
                    name=f'Generation {generation} Progress {progress * 100:.0f}%'
                )

                if progress == start_progress_limit or progress == end_progress_limit:
                    for i in range(15):
                        frames.append(transition_frame)
                frames.append(transition_frame)

        # Actualizar los frames de la figura
        fig.frames = frames

        # Mostrar la animación
        fig.show()

    def plot_evolution(self) -> None:

        data_dict = self.populuation_dict
        generations = sorted(data_dict.keys())  # Obtener las generaciones en orden

        # Preparar los datos para cada fotograma de la animación
        max_individuals = max(len(generation) for generation in data_dict.values())

        # Obtengo losfitness mas grandes y mas pequeños
        max_fitness_value: float = max(
            [ind.get_individual_fitness() for ind in [elemento for sublista in data_dict.values() for elemento in sublista] if ind.get_individual_fitness() is not None])
        min_fitness_value: float = min(
            [ind.get_individual_fitness() for ind in [elemento for sublista in data_dict.values() for elemento in sublista] if ind.get_individual_fitness() is not None])

        def generate_distinct_colors(n):
            colors = []
            for i in range(n):
                hue = i / n
                saturation = 0.7
                lightness = 0.5
                rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
                colors.append(f'rgb({int(rgb[0] * 255)},{int(rgb[1] * 255)},{int(rgb[2] * 255)})')
            return colors

        # Generar un color por generación
        generation_colors = generate_distinct_colors(len(generations))
        generation_color_map = dict(zip(generations, generation_colors))

        # Crear el layout inicial con individuos de color de su generación
        initial_data = []
        for generation in generations:
            population = data_dict[generation]
            gen_color = generation_color_map[generation]
            for i, ind in enumerate(population):
                initial_data.append(go.Scatter(
                    x=[i],
                    y=[ind.get_individual_fitness()],  # Usar el fitness real del individuo
                    mode='markers',
                    marker=dict(
                        color=gen_color,
                        size=14,
                        opacity=0.7,
                        line=dict(width=1.5, color='DarkSlateGrey')
                    ),
                    name=f'Gen {generation} Ind {i}',
                    text=[f'Gen: {generation}<br>Ind: {i}<br>Fitness: {ind.get_individual_fitness() if ind.get_individual_fitness() is not None else 0:.3f}'],
                    hoverinfo='text',
                    textposition='top center'
                ))

        # Configurar la figura con animación
        fig = go.Figure(
            data=initial_data,
            layout=go.Layout(
                # Título con estilo profesional
                title=dict(
                    text='Evolución de la Función de Fitness a lo largo de las generaciones',
                    font=dict(
                        size=22,
                        family='Arial, sans-serif',
                        color='#2C3E50'  # Azul profundo elegante
                    ),
                    x=0.5,
                    xanchor='center',
                    pad=dict(b=20, t=20)
                ),

                # Fondo profesional
                plot_bgcolor='rgba(245,248,250,0.9)',  # Gris azulado muy claro
                paper_bgcolor='white',

                xaxis=dict(
                    title='Individuos',
                    range=[-1, max_individuals],
                    tickmode='linear',
                    tick0=0,
                    dtick=1,
                    gridcolor='lightgrey',
                    zerolinecolor='lightgrey',
                    showline=True,
                    linewidth=2,
                    linecolor='#BDC3C7'
                ),
                yaxis=dict(
                    title='Fitness',
                    range=[-0.1 if min_fitness_value > -0.1 else min_fitness_value, 1.1 if max_fitness_value < 1.1 else max_fitness_value],
                    gridcolor='lightgrey',
                    zerolinecolor='lightgrey',
                    showline=False,
                    linewidth=2,
                    linecolor='#BDC3C7'
                ),

                # Leyenda con estilo
                legend=None,
                showlegend=False,
                hovermode='closest',
            ),
        )

        # Mostrar la animación
        fig.show()

    # </editor-fold>

__all__ = ['Population']



from typing import List, Literal
from aletheia_quantum_genetic_optimizers.individuals import Individual
import numpy as np
import random


class Mutation:
    def __init__(self, individual_list: List[Individual],
                 mutate_probability: float,
                 mutate_gen_probability: float,
                 mutation_policy: Literal['soft', 'normal', 'hard'],
                 problem_restrictions: Literal['bound_restricted', 'totally_restricted'],
                 num_generations: int):
        """
        Inicializa una instancia de la clase encargada de gestionar una población de individuos en un algoritmo evolutivo.

        Parámetros:
        - individual_list (List[Individual]): Lista de individuos que conforman la población inicial.
        - mutate_probability (float): Probabilidad de que un individuo sufra mutación en una generación.
        - mutate_gen_probability (float): Probabilidad de que un gen específico dentro de un individuo sea mutado.
        - mutation_policy (Literal['soft', 'normal', 'hard']): Define la severidad de la mutación aplicada.
        - problem_restrictions (Literal['bound_restricted', 'totally_restricted']): Tipo de restricción aplicada sobre los valores posibles de los genes.
        - num_generations (int): Número total de generaciones que el algoritmo debe ejecutar.
        """

        self.individual_list: List[Individual] = individual_list
        self.mutate_probability: float = mutate_probability
        self.mutate_gen_probability: float = mutate_gen_probability
        self.mutation_policy: Literal['soft', 'normal', 'hard'] = mutation_policy

        # -- Almaceno en una propiedad las restricciones a aplicar
        self.problem_restrictions: Literal['bound_restricted', 'totally_restricted'] = problem_restrictions
        self.bounds_dict = self.individual_list[0].bounds_dict
        self.num_generations: int = num_generations

    def run_mutation(self):
        match self.problem_restrictions:
            case "bound_restricted":
                return self.bound_restricted_mutation()
            case "totally_restricted":
                return self.totally_restricted_mutation()

    # <editor-fold desc="Mutaciones en funcion del self.problem_restrictions    --------------------------------------------------------------------------------------------------">

    def bound_restricted_mutation(self) -> List[Individual]:

        """
        Política de mutación para propiedades entre bounds
        :return: [List[Individual]] Lista de individuos
        """

        # -- Evaluamos la necesidad de mutar cada individuo de la lista de individuos de la generación
        for individual in self.individual_list:

            # -- Si no se debe mutar, entonces se continúa con el siguiente individuo
            if np.random.rand() >= self.mutate_probability:
                continue

            # -- Si se debe mutar el individuo...

            # -- Realizamos los cruces de cada gen
            for parameter, bound in self.bounds_dict.items():

                # -- Analizamos por tipo de bound
                match bound['bound_type']:

                    case 'predefined':

                        # -- Si debe mutar una propiedad | gen
                        if np.random.rand() < self.mutate_gen_probability:
                            mutated_parameter, mutated_angles = self.mutation_bit_flip(individual, parameter)
                            individual.set_individual_value(parameter, mutated_parameter)
                            individual.set_individual_angles(parameter, mutated_angles)

                    case 'interval':
                        if np.random.rand() < self.mutate_gen_probability:
                            mutated_parameter, mutated_angles = self.mutation_uniform(individual, parameter)
                            individual.set_individual_value(parameter, mutated_parameter)
                            individual.set_individual_angles(parameter, mutated_angles)

        return self.individual_list

    def totally_restricted_mutation(self):
        """Realiza una mutación por intercambio en la lista dada."""

        for individual in self.individual_list:
            if np.random.rand() >= self.mutate_probability:
                continue

            individual_values = [z for z in individual.get_individual_values().values()]

            if len(individual_values) < 2:
                return individual_values  # No se puede mutar si hay menos de 2 elementos
            i, j = random.sample(range(len(individual_values)), 2)  # Escoge dos índices distintos
            individual_values[i], individual_values[j] = individual_values[j], individual_values[i]  # Intercambia los valores

            individual.set_individual_values(individual_values)

        return self.individual_list

    # </editor-fold>

    # <editor-fold desc="Metodos de mutacion de genes    -------------------------------------------------------------------------------------------------------------------------">

    def mutation_bit_flip(self, individual: Individual, parameter: str):
        """
        Mutación de variables continuas dentro de un rango
        :param individual: [Individual] Individuo de la generación del cual se debe mutar una de sus propiedades|genes
        :param parameter: [str] Parámetro | Ángulos de rotación que se deben mutar
        :return:
        """
        total_generations = self.num_generations
        current_generation: int = individual.generation
        mutation_progress = 1 - (current_generation / total_generations)  # -- Reduce progresivamente

        # -- Definimos los límites de mutación en función de la política de mutación

        match self.mutation_policy:

            case 'soft':
                # -- Inicialmente tomamos el 40% del rango total, y lo vamos reduciendo con cada generación
                percentage_range = 0.4 * mutation_progress

            case 'normal':
                # -- Inicialmente tomamos el 60% del rango total, y lo vamos reduciendo con cada generación
                percentage_range = 0.6 * mutation_progress

            case 'hard' | _:
                # -- Mantenemos el rango completo sin reducción progresiva
                percentage_range = 0.8 * mutation_progress

        # -- Obtenemos los valores dentro de los límites permitidos
        possible_values = sorted(self.bounds_dict[parameter]["malformation_limits"])
        current_value = individual.get_individual_values()[parameter]

        # -- Determinamos el rango basado en la generación actual
        min_val, max_val = min(possible_values), max(possible_values)
        range_span = max_val - min_val
        lower_bound = max(min_val, current_value - (range_span * percentage_range))
        upper_bound = min(max_val, current_value + (range_span * percentage_range))

        # -- Filtramos valores dentro del rango calculado
        filtered_values = [z for z in possible_values if lower_bound <= z <= upper_bound and z != current_value]

        # -- Si no quedan valores posibles, tomar el más cercano válido
        if not filtered_values:
            filtered_values = [z for z in possible_values if z != current_value]

        # -- Mutamos las propiedades|hiperparámetros (numeros enteros|flotantes)
        mutated_parameter = float(np.random.choice(filtered_values)) if self.bounds_dict[parameter]["type"] == "float" else int(np.random.choice(filtered_values))

        # -- Calculamos el valor mutado de los ángulos en la escala [-pi, pi]

        # -- Aplicamos la proporción de cambio a cada ángulo, contemplando los cambios de signo
        modified_angles: List[float] = []

        for angle in individual.get_vqc_parameters_values()[parameter]:

            # -- Convertimos el valor actual a su escala angular correspondiente.
            current_angle = ((current_value - min_val) / range_span) * (2 * np.pi) - np.pi

            # -- Convertimos el valor mutado a su escala angular correspondiente.
            mutated_angle = ((mutated_parameter - min_val) / range_span) * (2 * np.pi) - np.pi

            # -- Calculamos la proporción de cambio entre ángulos.
            mutation_angle_change = mutated_angle - current_angle

            # -- Aplicamos la proporción de cambio al ángulo actual.
            modified_angle = angle + mutation_angle_change

            # -- Si la mutación cruza de positivo a negativo o viceversa, se debe reflejar en el ángulo.
            if (angle > 0 > modified_angle) or (angle < 0 < modified_angle):
                modified_angle = -modified_angle  # Invertimos el signo si cruza el cero

            # -- Normalizamos el ángulo al rango [-π, π].
            modified_angle = ((modified_angle + np.pi) % (2 * np.pi)) - np.pi

            # -- Agregamos el ángulo modificado a la lista de nuevos ángulos.
            modified_angles.append(float(modified_angle))

        return mutated_parameter, modified_angles

    def mutation_uniform(self, individual, parameter):
        """
        Realiza una mutación uniforme en valores enteros o reales.
        :param individual: Individuo que se quiere mutar en alguno de sus genes.
        :param parameter: Parámetro que se quiere modificar del individuo.

        :return: Parámetro mutado.
        """

        total_generations = self.num_generations
        current_generation: int = individual.generation
        mutation_progress = 1 - (current_generation / total_generations)  # Reduce progresivamente

        # -- Definimos los límites de mutación en función de la política
        match self.mutation_policy:

            case 'soft':
                percentage_range = 0.4 * mutation_progress

            case 'normal':
                percentage_range = 0.6 * mutation_progress

            case 'hard' | _:
                percentage_range = 0.8 * mutation_progress

        # -- Obtenemos los valores dentro de los límites permitidos
        possible_values = sorted(self.bounds_dict[parameter]["malformation_limits"])
        current_value = individual.get_individual_values()[parameter]

        # -- Determinamos el rango basado en la generación actual
        min_val, max_val = min(possible_values), max(possible_values)
        range_span = max_val - min_val
        lower_bound = max(min_val, current_value - (range_span * percentage_range))
        upper_bound = min(max_val, current_value + (range_span * percentage_range))

        # -- Ajustamos los límites para valores flotantes o enteros

        # -- Aseguramos que el valor mutado esté dentro de los límites definidos
        if self.bounds_dict[parameter]["type"] == "float":
            mutated_parameter = float(np.random.uniform(lower_bound, upper_bound))
        else:
            mutated_parameter = int(np.random.uniform(np.floor(lower_bound), np.ceil(upper_bound)))

        # -- Control adicional para evitar valores extremos no deseados
        mutated_parameter = int(np.clip(mutated_parameter, min_val, max_val)) if isinstance(mutated_parameter, int) else float(np.clip(mutated_parameter, min_val, max_val))

        # -- Modificamos los ángulos en la misma proporción
        modified_angles: List[float] = []

        for angle in individual.get_vqc_parameters_values()[parameter]:

            # -- Convertimos el valor actual a su escala angular correspondiente.
            current_angle = ((current_value - min_val) / range_span) * (2 * np.pi) - np.pi

            # -- Convertimos el valor mutado a su escala angular correspondiente.
            mutated_angle = ((mutated_parameter - min_val) / range_span) * (2 * np.pi) - np.pi

            # -- Calculamos la proporción de cambio entre ángulos.
            mutation_angle_change = mutated_angle - current_angle

            # -- Aplicamos la proporción de cambio al ángulo actual.
            modified_angle = angle + mutation_angle_change

            # -- Si la mutación cruza de positivo a negativo o viceversa, se debe reflejar en el ángulo.
            if (angle > 0 > modified_angle) or (angle < 0 < modified_angle):
                modified_angle = -modified_angle  # Invertimos el signo si cruza el cero

            # -- Normalizamos el ángulo al rango [-π, π].
            modified_angle = ((modified_angle + np.pi) % (2 * np.pi)) - np.pi

            # -- Agregamos el ángulo modificado a la lista de nuevos ángulos.
            modified_angles.append(float(modified_angle))

        return mutated_parameter, modified_angles

    # </editor-fold>

__all__ = ['Mutation']


from aletheia_quantum_genetic_optimizers.quantum_technology import QuantumTechnology
from aletheia_quantum_genetic_optimizers.individuals import Individual
from singleton_tools import SingletonMeta

from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit.circuit.parametertable import ParameterView
from qiskit.circuit import ParameterVector, Parameter
from plotly.subplots import make_subplots
from typing import List, Literal, Dict
from itertools import combinations
from info_tools import InfoTools
from random import choices

import plotly.graph_objects as go
import matplotlib.pyplot as plt
import plotly.io as pio
import pandas as pd
import numpy as np
import itertools
import random

class Graph(metaclass=SingletonMeta):

    def __init__(self):

        self.graph: bool = True


class Reproduction:
    def __init__(self, winners_list: List[Individual],
                 number_of_children: int,
                 problem_restrictions: Literal['bound_restricted', 'totally_restricted'],
                 return_to_origin:[Literal['return_to_origin', 'no_return'] | None],
                 problem_type: Literal["minimize", "maximize"],
                 metric_to_optimize: Literal['accuracy', 'recall', 'specificity', 'f1', 'aur_roc', 'precision', 'negative_precision', 'mae', 'mse', 'other', 'r2'],
                 verbose: bool = True):

        """
        Clase | Objeto de reproducción de individuos
        :param winners_list: List[Individual] Lista de ganadores.
        :param number_of_children: [int] Cantidad de individuos que se quieren reproducir
        :param problem_restrictions: Literal['bound_restricted', 'totally_restricted'] Tipo de problema en cuanto a reestricciones
        :param return_to_origin:[Literal['return_to_origin', 'no_return'] | None] En caso de problemas totally_restricted es necesario saber si el problema termina en el origen o no es necesario que suceda esto
        :param problem_type: Literal["minimize", "maximize"] Tipo de problema en cuanto a maximización o minimización del resultado de la función objetivo
        :param metric_to_optimize: Literal['accuracy', 'recall', 'specificity', 'f1', 'aur_roc', 'precision', 'negative_precision', 'mae', 'mse', 'other'] Métrica a optimizar en la función objetivo
        :param verbose: [bool] Variable para definir si se pinta en consola | grafica
        """

        # -- Definimos las variables generales
        self.winners_list: List[Individual] = winners_list
        self.number_of_children: int = number_of_children
        self.problem_restrictions: Literal['bound_restricted', 'totally_restricted'] = problem_restrictions
        self.return_to_origin: Literal['return_to_origin', 'no_return'] | None = return_to_origin
        self.children_list: List[Individual] = []
        self.parents_generation: int = self.winners_list[0].generation
        self.problem_type: Literal['minimize', 'maximize'] = problem_type
        self.metric_to_optimize: Literal['accuracy', 'recall', 'specificity', 'f1', 'aur_roc', 'precision', 'negative_precision', 'mae', 'mse', 'other', 'r2'] = metric_to_optimize
        self.verbose: bool = verbose

        # -- Instanciamos clases de apoyo
        self.IT: InfoTools = InfoTools()
        self.QT: QuantumTechnology | None = None

        # -- Añadimos variables que luego determinaremos en run_reproduction
        self.optimization_quantum_technology: Literal["simulator", "quantum_machine"] | None = None
        self.optimization_service: Literal["aer", "ibm"] | None = None
        self.qm_api_key: str | None = None
        self.qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None = None
        self.quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"] | None = None
        self.generations_fitness_statistics_df: pd.DataFrame | None = None
        self.max_qubits: int | None = None

        # -- Añadimos un dataset para guardar los resultados de las normalizaciones entre binario, pi y bounds_dict
        self._reescaling_result_df: pd.DataFrame | None = None

        # -- Instanciamos la clase Graph (Singleton para evitar graficación excesiva con verbose True)
        self.GPH: Graph = Graph()

    def run_reproduction(self, quantum_technology: Literal["simulator", "quantum_machine"], service: Literal["aer", "ibm"], qm_api_key: str | None, qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None, quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"], generations_fitness_statistics_df: pd.DataFrame, max_qubits: int):
        """
        Metodo para ejecutar la reproducción de los mejores individuos
        :param quantum_technology. Literal["simulator", "quantum_machine"]. Tecnología cuántica con la que calculan la lógica. Si es simulator, se hará con un simulador definido en el
        parámetro service. Si es quantum_machine, el algoritmo se ejecutará en una máquina cuántica definida en el parámetro technology.
        :param service. ["aer", "ibm"] El servicio tecnológico con el cual se ejecuta la lógica.
        :param qm_api_key. [str] API KEY para conectarse con el servicio de computación cuántica de una empresa.
        :param qm_connection_service. Literal["ibm_quantum", "ibm_cloud"] | None. Servicio específico de computación cuántica. Por ejemplo, en el caso de IBM pueden ser a la fecha ibm_quantum | ibm_cloud
        :param quantum_machine. Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"]. Nombre del ordenador cuántico a utilizar. Por ejemplo, en el caso de IBM puede ser ibm_brisbane, ibm_kyiv, ibm_sherbrooke. Si se deja en least_busy,
        se buscará el ordenador menos ocupado para llevar a cabo la ejecución del algoritmo cuántico.
        :param generations_fitness_statistics_df: [pd.Dataframe] Df que contiene la información estadística de todas.
        :param max_qubits: [int | None] Número máximo de qubits a utilizar.
        :return: La siguiente generación de individuos
        """

        self.optimization_quantum_technology: Literal["simulator", "quantum_machine"] = quantum_technology
        self.optimization_service: Literal["aer", "ibm"] = service
        self.qm_api_key: str | None = qm_api_key
        self.qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None = qm_connection_service
        self.quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"] = quantum_machine
        self.generations_fitness_statistics_df: pd.DataFrame = generations_fitness_statistics_df
        self.max_qubits: int | None = max_qubits

        self.QT: QuantumTechnology = QuantumTechnology(quantum_technology=self.optimization_quantum_technology,
                                                       service=self.optimization_service,
                                                       qm_api_key=self.qm_api_key,
                                                       qm_connection_service=self.qm_connection_service,
                                                       quantum_machine=self.quantum_machine)

        if self.verbose:
            self.IT.intro_print(f"RUN REPRODUCTION: La generacion {self.parents_generation} reproducirá la generación {self.parents_generation + 1}")

        # -- Agregamos a los 10 % de mejores padres no repetidos a la children list
        num_to_select = max(1, int(len(self.winners_list) * 0.15))

        # -- Seleccionamos los mejores individuos según el tipo de problema (maximize o minimize)
        sorted_winners = sorted(self.winners_list,
                                key=lambda ind: ind.get_individual_fitness(),
                                reverse=self.problem_type == "maximize")

        # -- Almacenamos los mejores individuos
        selected_individuals = sorted_winners[:num_to_select]

        # -- Añadimos estos individuos a la children_list
        self.children_list += selected_individuals

        # -- Después, creamos individuos idénticos a esos padres porque no podemos modificar su generación sin dejar descalza la generacion 0
        self.children_list = [
            Individual(
                bounds_dict=ind.bounds_dict,
                child_values=ind.get_child_values(),
                vqc=self.winners_list[0].get_vqc(),
                vqc_parameters=ind.get_vqc_parameters_values(),
                generation=self.parents_generation + 1,
                problem_restrictions=self.problem_restrictions,
                element_matrix=self.winners_list[0].get_element_matrix())
            for ind in self.children_list]

        # -- Tenemos en cuenta las restricciones del problema
        match self.problem_restrictions:
            case "bound_restricted":
                return self.bound_restricted_reproduction()
            case "totally_restricted":
                return self.totally_restricted_reproduction()

    def bound_restricted_reproduction(self):

        # -- Obtenemos el bounds_dict para que cada tipo de parametro tenga su propio cruce
        bounds_dict = self.winners_list[0].bounds_dict

        # -- Obtenemos la combinación posible de individuos ganadores
        combinations_of_individuals = list(itertools.permutations(self.winners_list, r=2))

        # -- Creamos un diccionario para almacenar los circuitos cuánticos de cada combinación y propiedad
        combo_circuits: dict = {}

        # -- Para cada una de las combinaciones de individuos
        for idx, combo in enumerate(combinations_of_individuals):

            combo_circuits[idx] = {}

            # -- Obtenemos los ids de los individuos
            ind1: int = combo[0].get_individual_id()
            ind2: int = combo[1].get_individual_id()

            if self.verbose:
                self.IT.sub_intro_print("Reproducción entre individuos")
                self.IT.info_print(f"Reproducción entre individuo {ind1} e individuo {ind2}")

            # -- Por cada par de individuos reconstruimos sus respectivos circuitos cuánticos y quitamos medidas
            ind1_vqc: QuantumCircuit = combo[0].get_vqc().copy()
            ind1_vqc.remove_final_measurements()

            ind2_vqc: QuantumCircuit = combo[1].get_vqc().copy()
            ind2_vqc.remove_final_measurements()

            # -- Agregamos a los circuitos nuevos parámetros para las métricas
            ind1_vqc = self.QT.add_new_qubit(ind1_vqc, Parameter("metric_0"), combo[0].bounds_dict, max_qubits=self.max_qubits)
            ind2_vqc = self.QT.add_new_qubit(ind2_vqc, Parameter("metric_1"), combo[1].bounds_dict, max_qubits=self.max_qubits)

            # -- Obtenemos los parámetros de los circuitos
            parameters: dict = {"0":ind1_vqc.parameters, "1":ind2_vqc.parameters}

            # -- Creamos un diccionario para almacenar los parámetros de los circuitos para luego asignarlos
            parent_circuits: dict = {}

            # -- Iteramos por cada padre
            for i in range(0, 2):

                # -- Normalizamos el resultado de la función objetivo de cada padre
                metric_to_optimize: dict = {f"{combo[i].get_individual_id()}": self.QT.normalize_metric(
                    generations_fitness_statistics_df=self.generations_fitness_statistics_df,
                    metric_name=self.metric_to_optimize,
                    value=combo[i].get_individual_fitness())}

                # -- Obtenemos los parametros theta de los qubits ("Y" y "Z")
                bound_angles = [angle for parameter in combo[0].bounds_dict.keys() for angle in
                                combo[i].get_vqc_parameters_values()[parameter]]

                # -- Añadimos los ángulos de los parámetros (incluido el valor de la función objetivo)
                parameters_dict: dict = {
                    param: value for param, value in zip([z for z in parameters[str(i)]],
                                                         [z for z in metric_to_optimize.values()] +
                                                         # [z / 2 for z in metric_to_optimize.values()] +
                                                         [z for z in bound_angles])
                }

                # -- Guardamos los parámetros de cada circuito
                parent_circuits[i] = {"parameters_dict": parameters_dict}

            # -- Asignamos los valores de las métricas
            ind1_vqc = ind1_vqc.assign_parameters(parent_circuits[0]["parameters_dict"]).copy()
            ind2_vqc = ind2_vqc.assign_parameters(parent_circuits[1]["parameters_dict"]).copy()

            # -- Creamos un circuito nuevo a partir de los padres y agregamos una puerta CNOT al final
            combined_qc: QuantumCircuit = self.QT.create_parent_vqc(bounds_dict, ind1_vqc, ind2_vqc).copy()

            if self.verbose and self.GPH.graph is True:
                self.QT.advanced_qc_plot(combined_qc)


            # -- Añadimos las mediciones al circuito cuántico parental creado (procedimientos especial):
            # -- 1. Multiplicamos los circuitos: Generamos una réplica por cada propiedad.
            # -- 2. Agregamos mediciones: En todos los qubits excepto en aquellos de los cuales no queremos propiedades.
            # -- Por ejemplo: Si queremos ejecutar un circuito para obtener el número binario que luego reescalaremos en
            # -- el rango -pi y pi, y luego entre los bounds, en particular para max_depth, entonces agregamos medidas
            # -- para todos los qubits, excepto para los qubits de otras propiedades (ej: n_estimators del padre 1 y 2).

            # -- Añadimos las medidas a las vqc concatenadas de los mejores padres
            combined_qc: dict = self.QT.adding_measurements_parent_vqc(bounds_dict=bounds_dict,
                                                                       combined_qc=combined_qc,
                                                                       verbose=self.verbose,
                                                                       graph=self.GPH.graph).copy()

            # -- Hacemos que no se pueda graficar de vuelta (evitar sobreproducción de imágenes)
            self.GPH.graph = False

            # -- Guardamos el circuito con las medidas en el diccionario de circuitos parentales
            for parameter, circuit in combined_qc.items():
                combo_circuits[idx][parameter] = circuit

            # -- Guardamos los ángulos/parámetros de los padres que generaron estos circuitos
            combo_circuits[idx]["individuals"] = {"individual_0": combo[0], "individual_1": combo[1]}

        # -- Almacenamos en un diccionario las listas de circuitos que pueden generar cada una de las propiedades
        prop_vqc_dict: dict = {}
        for idx, sub_dict in combo_circuits.items():
            for k, v in sub_dict.items():
                if k in [z for z in {key for d in combo_circuits.values() for key in d.keys()} if 'individuals' not in z]:
                    if k not in prop_vqc_dict.keys():
                        prop_vqc_dict[k] = [v]
                    else:
                        prop_vqc_dict[k].append(v)

        # -- Eliminamos la cantidad de circuitos que exceden el numero de hijos a reproducir
        first_key = next(iter(prop_vqc_dict))
        excess = len(prop_vqc_dict[first_key]) - self.number_of_children + len(self.children_list)

        if excess > 0:

            # -- Generamos índices aleatorios a eliminar
            indices_to_remove = sorted(random.sample(range(len(prop_vqc_dict[first_key])), excess), reverse=True)

            # -- Eliminamos esos índices de cada lista
            for k in prop_vqc_dict:
                for idx in indices_to_remove:
                    del prop_vqc_dict[k][idx]

        # -- Ejecutamos cada uno de los circuitos 20k veces
        results_dict: dict = {}
        for indv_prop, circuits in prop_vqc_dict.items():
            results_dict[indv_prop] = self.QT.execution_object.run(circuits, shots=20000)

        # -- TODO: Se podría implementar el código comentado y comentar el de abajo
        #  Nos quedamos por cada individuo con los 2 valores binarios que salieron más veces para cada propiedad
        """individual_dict: dict = {}
        for k, v in results_dict.items():
            for i, shot in enumerate(v):
                if i in [z for z in individual_dict.keys()]:
                    individual_dict[i]["binary"] = individual_dict[i]["binary"] | {k: [z for z in dict(sorted(shot.items(), key=lambda x: x[1], reverse=True)[:1]).keys()]}
                    individual_dict[i]["pi"] = individual_dict[i]["pi"] | {k: [self.QT.binary_to_float_and_normalize_pi([z]) for z in individual_dict[i]["binary"][k]]}
                    individual_dict[i]["real"] = individual_dict[i]["real"] | {k: [self.QT.rescaling_pi_to_integer_float(bounds_dict, k, individual_dict[i]["pi"][k][0])][0]}
                else:
                    individual_dict[i] = {"binary": {k: [z for z in dict(sorted(shot.items(), key=lambda x: x[1], reverse=True)[:1]).keys()]}}
                    individual_dict[i] = individual_dict[i] | {"pi": {k: [self.QT.binary_to_float_and_normalize_pi([z]) for z in individual_dict[i]["binary"][k]]}}
                    individual_dict[i] = individual_dict[i] | {"real": {k: [self.QT.rescaling_pi_to_integer_float(bounds_dict, k, individual_dict[i]["pi"][k][0])][0]}}"""

        individual_dict: dict = {}
        for k, v in results_dict.items():
            for i, shot in enumerate(v):

                if i in [z for z in individual_dict.keys()]:
                    individual_dict[i]["binary"] = individual_dict[i]["binary"] | {k: random.choice(list(shot.keys()))}
                    individual_dict[i]["pi"] = individual_dict[i]["pi"] | {k: [self.QT.binary_to_float_and_normalize_pi([individual_dict[i]["binary"][k]])]}
                    individual_dict[i]["real"] = individual_dict[i]["real"] | {k: [self.QT.rescaling_pi_to_integer_float(bounds_dict, k, individual_dict[i]["pi"][k][0])][0]}
                else:
                    individual_dict[i] = {"binary": {k: random.choice(list(shot.keys()))}}
                    individual_dict[i] = individual_dict[i] | {"pi": {k: [self.QT.binary_to_float_and_normalize_pi([individual_dict[i]["binary"][k]])]}}
                    individual_dict[i] = individual_dict[i] | {"real": {k: [self.QT.rescaling_pi_to_integer_float(bounds_dict, k, individual_dict[i]["pi"][k][0])][0]}}

        # -- Generamos un daframe con los resultados (binarios, angulos y propiedades)
        if self._reescaling_result_df is None:
            individual_dict_df: pd.DataFrame = pd.DataFrame(individual_dict).T
            individual_dict_df["parent_generation"] = self.winners_list[0].generation
            individual_dict_df["children_generation"] = self.winners_list[0].generation + 1
            self._reescaling_result_df = individual_dict_df
        else:
            individual_dict_df: pd.DataFrame = pd.DataFrame(individual_dict).T
            individual_dict_df["parent_generation"] = self.winners_list[0].generation
            individual_dict_df["children_generation"] = self.winners_list[0].generation + 1
            self._reescaling_result_df = pd.concat([self._reescaling_result_df, individual_dict_df])

        if self.verbose:
            self.IT.print_tabulate_df(self._reescaling_result_df, row_print=100, show_index=True)

        for i, row in self._reescaling_result_df[self._reescaling_result_df["parent_generation"] == self.winners_list[0].generation].iterrows():
            self.children_list.append(Individual(bounds_dict=bounds_dict,
                                                child_values=[z for z in row["real"].values()],
                                                vqc=self.winners_list[0].get_vqc(),
                                                vqc_parameters=row["pi"],
                                                generation=row["children_generation"],
                                                problem_restrictions=self.winners_list[0].problem_restrictions))

        for i in self.children_list:
            self.IT.info_print(f"INDIVIDUO {i.get_individual_id()} --> PROPIEDADES: {i.get_individual_values()} --> PARAMETROS: {i.get_vqc_parameters_values()}")

        if self.verbose:
            for i in self.children_list:
                self.IT.info_print(f"INDIVIDUO {i.get_individual_id()} --> PROPIEDADES: {i.get_individual_values()} --> PARAMETROS: {i.get_vqc_parameters_values()}")

        return self.children_list

    def totally_restricted_reproduction(self):

        self.IT.header_print("Iniciamos el proceso de reproducción de los individuos")

        # -- Obtenemos el bounds_dict para el momento de reproducir un individuo nuevo
        bounds_dict: Dict = self.winners_list[0].bounds_dict

        # ----------------------------------------------------------------------------------------
        # -- Separamos en clústers los individuos a fin de hacer computacionalmente viable el QAOA
        # ----------------------------------------------------------------------------------------

        # -- Obtenemos el número de elementos que conforman el individuo (Ej. ciudades del problema del TSP)
        num_elements: int = len([z for z in self.winners_list[0].get_element_matrix().keys()])

        # -- Definimos variables que utilizaremos en la generación del problema del QAOA
        zero_cluster: int | None = None
        contains_zero: bool = False

        # -- Iteramos el proceso de creación de individuos sin malformaciones hasta completar el número de hijos
        while len(self.children_list) < self.number_of_children:

            # -- Seleccionamos otro individuo al azar con el que se va a cruzar
            random_individual_selected_1: Individual = random.choice([ind for ind in self.winners_list])
            random_individual_selected_2: Individual = random.choice([ind for ind in self.winners_list if ind != random_individual_selected_1.get_individual_values()])

            if self.verbose:
                self.IT.sub_intro_print(f"Padres que se van a reproducir:")
                self.IT.info_print(f"Padre 1: {random_individual_selected_1.get_individual_values()}")
                self.IT.info_print(f"Padre 2: {random_individual_selected_2.get_individual_values()}")

            # -- Realizamos los cruces entre dos padres para sacar un hijo
            c1: List = self.ox1([z for z in random_individual_selected_1.get_individual_values().values()],
                          [z for z in random_individual_selected_2.get_individual_values().values()])

            # -- Separamos en clústeres [centroids no se usan parece]
            clusters = self.divide_ids_into_clusters_with_coords(
                ordered_ids=c1,
                id_to_coords=self.winners_list[0].get_element_matrix(),
                min_size=2,
                max_size=5
            )

            if self.verbose:
                self.IT.info_print(f"Se han creado {len([z for z in clusters])} clústers para {num_elements} elementos")

            self.IT.intro_print(f"Reproduciendo el individuo {len(self.children_list) + 1}")

            self.IT.sub_intro_print("Clústers")

            for cid, c in clusters.items():
                self.IT.info_print(f"Clúster {cid}: {list(c.keys())}")

            # -- Si el problema es de tipo return_to_origin == no_return identificamos el clúster con el primer elemento
            if self.return_to_origin == "no_return":

                # -- Identificamos el clúster que contiene el primer elemento (ej. la primera ciudad del TSP)
                zero_cluster: int = next(cid for cid, cluster in clusters.items() if 0 in cluster)

                if self.verbose:
                    self.IT.info_print(f"Modo no_return: El clúster con el primer elemento es el {zero_cluster}")

            # -- Definimos la lista con las combinaciones de elementos
            cluster_combinations: list = []

            # -- Iteramos por cada clúster
            for cid, cluster_element_matrix in clusters.items():
                route, dist, _ = None, None, None

                self.IT.sub_intro_print(f"Optimizando clúster {cid} --> {cluster_element_matrix}")

                if self.return_to_origin == "no_return":

                    # -- El primer clúster debe empezar desde la ciudad 0
                    contains_zero = cid == zero_cluster

                # -- Intentamos resolver el problema con el QAOA
                self.IT.info_print(f"Resolviendo el problema con QAOA")
                combination = self.QT.solve_tsp_with_qaoa(
                    element_matrix=cluster_element_matrix,
                    problem_type=self.problem_type,
                    contains_zero=contains_zero,
                    return_to_origin=self.return_to_origin,
                )

                # -- Si existe una combinación completa del QAOA y posee todos los elementos, la agregamos a la lista
                if combination and len(combination) == len([z for z in cluster_element_matrix.keys()]):
                    cluster_combinations.append(combination)
                    self.IT.info_print("Se ha agregado una combinación completa a la lista")

                # -- Si no existe una combinación completa del QAOA o no posee todos los elementos, agregamos una random
                if not combination or len(combination) != len([z for z in cluster_element_matrix.keys()]):
                    self.IT.info_print(f"[WARNING] No se encontró combinación válida en cluster. Generando ruta aleatoria...", "light_yellow")
                    combinations = list(cluster_element_matrix.keys())
                    random_combination = random.sample(combinations, len(combinations))

                    self.IT.info_print("Se ha agregado una combinación aleatoria completa a la lista", "light_yellow")
                    cluster_combinations.append(random_combination)

            # -- Combinamos todas las rutas de los clústers
            final_combination: list = [item for sublist in cluster_combinations for item in sublist]

            self.IT.info_print(f"La combinacion final de elementos es: {final_combination}")

            if self.verbose:
                self.plot_tsp_route(element_matrix=self.winners_list[0].get_element_matrix(),
                                    combination=final_combination,
                                    return_to_origin=self.return_to_origin)

            # -- Verificamos si la ruta final contiene todas las ciudades
            if len(final_combination) == len([z for z in self.winners_list[0].get_element_matrix().keys()]):
                # -- Crear el nuevo individuo con la ruta combinada
                new_child = Individual(bounds_dict=bounds_dict,
                                       child_values=final_combination,
                                       vqc=None,
                                       vqc_parameters=None,
                                       generation=self.winners_list[0].generation + 1,
                                       problem_restrictions=self.winners_list[0].problem_restrictions,
                                       element_matrix=self.winners_list[0].get_element_matrix())

                # -- Validamos que no existan individuos muy similares en la lista y que no tengan malformacion
                is_duplicate = any(existing_indv == new_child for existing_indv in self.children_list)
                if not is_duplicate and not new_child.get_individual_malformation():
                    self.children_list.append(new_child)
                    self.IT.info_print("Un nuevo individuo ha sido añadido a la lista con exito")

        # -- Mostramos información de los hijos
        for i in self.children_list:
            self.IT.info_print(
                f"INDIVIDUO {i.get_individual_id()} --> PROPIEDADES: {i.get_individual_values()} --> COMBINACION: {i.get_child_values()}")

        return self.children_list

    @staticmethod
    def ox1(parent1, parent2):
        """Aplica Order Crossover (OX1) a dos padres de un problema basado en permutaciones."""

        # -- Creamos una lista del tamaño de los genes y la inicializamos en None
        size = len(parent1)
        offspring = [None] * size

        # Seleccionamos un segmento aleatorio en el rango permitido, por ejemplo de [0,1,2,3,4,5] se selecciona [2,3,4]
        start, end = sorted(random.sample(range(size), 2))

        # Copiar segmento del primer padre
        offspring[start:end + 1] = parent1[start:end + 1]

        # Completar con genes del segundo padre en el mismo orden, evitando repeticiones
        p2_genes = [gene for gene in parent2 if gene not in offspring]

        # Rellenar los espacios vacíos manteniendo el orden
        idx = 0
        for i in range(size):
            if offspring[i] is None:
                offspring[i] = p2_genes[idx]
                idx += 1

        return offspring

    @staticmethod
    def divide_ids_into_clusters_with_coords(ordered_ids: List, id_to_coords: Dict, min_size: int = 2, max_size: int = 5):
        """
        Divide una lista de IDs en grupos de entre min_size y max_size,
        mapeando cada ID a sus coordenadas desde un diccionario dado.

        :param ordered_ids: Lista ordenada de IDs a agrupar
        :param id_to_coords: Diccionario con {id: (x, y)}
        :param min_size: Tamaño mínimo de grupo
        :param max_size: Tamaño máximo de grupo
        :return: Diccionario con clusters numerados y coordenadas mapeadas
        """
        total = len(ordered_ids)
        i = 0
        cluster_idx = 0
        clusters = {}

        while i < total:
            restante = total - i

            if min_size <= restante <= max_size:
                ids = ordered_ids[i:]
                clusters[cluster_idx] = {idx: id_to_coords[idx] for idx in ids}
                break

            if restante < min_size:
                if clusters:
                    for idx in ordered_ids[i:]:
                        clusters[cluster_idx - 1][idx] = id_to_coords[idx]
                else:
                    clusters[cluster_idx] = {idx: id_to_coords[idx] for idx in ordered_ids[i:]}
                break

            for tam in range(max_size, min_size - 1, -1):
                if (restante - tam) >= min_size or (restante - tam) == 0:
                    ids = ordered_ids[i:i + tam]
                    clusters[cluster_idx] = {idx: id_to_coords[idx] for idx in ids}
                    i += tam
                    cluster_idx += 1
                    break

        return clusters

    @staticmethod
    def plot_tsp_route(element_matrix: Dict[str, tuple], combination: List[int], return_to_origin: Literal['return_to_origin', 'no_return'] | None, title: str = "Ruta TSP"):
        """
        Gráfico para graficar combinaciones:
        - Fondo negro
        - Flechas de conexión
        - Pantalla completa
        - Combinación mostrada debajo del gráfico
        """

        pio.renderers.default = "browser"  # o 'notebook' si estás en Jupyter
        combination_copy = combination.copy()

        # -- Coordenadas
        x_coords = [element_matrix[node][0] for node in combination_copy]
        y_coords = [element_matrix[node][1] for node in combination_copy]

        # -- Cierre visual del ciclo si aplica

        if return_to_origin == "return_to_origin":
            if combination_copy[0] != combination_copy[-1]:
                x_coords.append(element_matrix[combination_copy[0]][0])
                y_coords.append(element_matrix[combination_copy[0]][1])
                combination_copy.append(combination_copy[0])

        # -- Subgráfico: uno para el gráfico, otro para la combinación
        fig = make_subplots(
            rows=2, cols=1,
            row_heights=[0.85, 0.15],
            vertical_spacing=0.05,
            specs=[[{"type": "scatter"}], [{"type": "table"}]]
        )

        # -- Flechas entre puntos
        for i in range(len(combination_copy) - 1):
            x0, y0 = element_matrix[combination_copy[i]]
            x1, y1 = element_matrix[combination_copy[i + 1]]
            fig.add_annotation(
                x=x1, y=y1, ax=x0, ay=y0,
                xref="x1", yref="y1", axref="x1", ayref="y1",
                showarrow=True,
                arrowhead=3,
                arrowsize=1,
                arrowwidth=2,
                arrowcolor="cyan"
            )

        # -- Ruta
        fig.add_trace(go.Scatter(
            x=x_coords, y=y_coords,
            mode='lines+markers',
            line=dict(color='cyan', width=2),
            marker=dict(size=8, color='white'),
            name='Ruta'
        ), row=1, col=1)

        # -- Nodos
        for node in combination_copy:
            x, y = element_matrix[node]
            fig.add_trace(go.Scatter(
                x=[x], y=[y],
                mode='markers+text',
                text=[str(node)],
                textfont=dict(color='white'),
                textposition="top center",
                marker=dict(size=10, color='orange'),
                name=f'Nodo {node}',
                showlegend=False
            ), row=1, col=1)

        # -- Tabla con la combinación
        fig.add_trace(go.Table(
            header=dict(values=["Combinación de nodos"],
                        fill_color='black',
                        font=dict(color='white', size=14),
                        align='left'),
            cells=dict(values=[[str(combination_copy)]],
                       fill_color='black',
                       font=dict(color='lightgreen', size=12),
                       align='left')
        ), row=2, col=1)

        # -- Estilo general
        fig.update_layout(
            title=title,
            plot_bgcolor="black",
            paper_bgcolor="black",
            font=dict(color="white"),
            margin=dict(l=20, r=20, t=60, b=20),
            showlegend=False
        )

        fig.update_xaxes(showgrid=False, zeroline=False, color='white', row=1, col=1)
        fig.update_yaxes(showgrid=False, zeroline=False, color='white', row=1, col=1)

        fig.show()

__all__ = ['Graph', 'Reproduction']


